from typing import Optional
from betconnect import enums
from betconnect import config


class BetConnectException(Exception):

    """
    base exception class

    """

    pass


class InValidPersonalisedProductionURL(BetConnectException):
    def __init__(self, url: str):
        """
        Raise when the supplied personalised betconnect url does not match the required pattern
        :param url: The supplied URL
        """
        super(InValidPersonalisedProductionURL, self).__init__(url)
        self.url = url

    def __str__(self) -> str:
        return (
            f"The personalised url supplied ({self.url}), "
            f"is not in the correct format. Expected ending in .betconnect.com/"
        )


class UnknownBetConnectEnvironment(BetConnectException):
    def __init__(self, environment: enums.Environment):
        """
        Raise when the environment supplied is not recognised as one available to end users. STAGING OR PRODUCTION.
        :param environment: The supplied environment value
        """
        super(UnknownBetConnectEnvironment, self).__init__(environment)
        self.environment = environment

    def __str__(self) -> str:
        return f"Expect environment in [{enums.Environment.PRODUCTION},{enums.Environment.STAGING}], got {self.environment}"


class LoginMissingTokenInResponse(BetConnectException):
    """
    Raise when the login response contains no login token.
    """

    def __str__(self) -> str:
        return f"Missing login token in response data"


class FailedLogin(BetConnectException):
    def __init__(self, username: str):
        """
        Raised when the user supplies the incorrect username and password
        :param username: the client supplied username
        """
        super(FailedLogin, self).__init__(username)
        self.username = username

    def __str__(self) -> str:
        return f"Failed Login for the username {self.username}. Please check both your username and password."


class UnexpectedResponseStatusCode(BetConnectException):
    def __init__(self, status_code: int, url: str):
        """
        Raised when an unexpected status code is returned from a response
        :param status_code: the response status code
        :param url: the request url
        """
        super(UnexpectedResponseStatusCode, self).__init__(status_code, url)
        self.url = url
        self.status_code = status_code

    def __str__(self) -> str:
        return f"Unexpected status code ({self.status_code}) returned for request: {self.url} "


class GamStopException(BetConnectException):
    """
    Account perferences has highlighted that you have a Gamstop against your username.
    API Access is forbidden for any user with a Gamstop ban against their username.
    If this is incorrect please contact your account manager.
    """

    def __str__(self) -> str:
        return "API access if strictly forbidden for users with a Gamstop ban against their account."


class MissingUserPerferences(BetConnectException):
    """
    Called when trying to access user_id property for my_bets but the account_preferences are unavailable.
    """

    def __str__(self) -> str:
        return "Unable to get user_id via get_user_preferences. Please supply a value or speak to your account manager"


class MinOddException(BetConnectException):
    """
    Exception raised when min odds requirements not met
    """

    def __init__(self, min_odds: float):
        """
        Raised when the user supplies an odd less than the site minimum
        :param min_odds:
        """
        super(MinOddException, self).__init__(min_odds)
        self.min_odds = min_odds

    def __repr__(self) -> str:
        return f"The supplied odds {self.min_odds} is less then the site minimum {config.BET_REQUEST_MIN_ODDS} "


class BetRequestIDFormatException(BetConnectException):
    """
    Exception raised when the format does not match a valid UUID format.
    i.e. 'c9bf9e57-1685-4c89-bafb-ff5af830be8a'
    """

    def __init__(self, bet_request_id: str):
        super(BetRequestIDFormatException, self).__init__(bet_request_id)
        self.bet_request_id = bet_request_id

    def __str__(self) -> str:
        return f"bet_request_id ({self.bet_request_id}) should be of length 36 and be a UUID"


class BetRequestIDStakeSizeException(BetConnectException):
    """
    raised when an invalid stake size is used. Should be an integer and multiples of £5.0
    """

    def __init__(self, stake_size: int):
        super(BetRequestIDStakeSizeException, self).__init__(stake_size)
        self.stake_size = stake_size

    def __str__(self) -> str:
        return f"Supplied stake size {self.stake_size} is not greater than the site minimum of {config.SITE_MINIMUM_STAKE_SIZE}"


class BetRequestInvalidCustomerOrderRefFormatException(BetConnectException):
    """
    raised when customer order ref is not between 1-36 character in length
    """

    def __init__(self, customer_order_ref: str):
        super(BetRequestInvalidCustomerOrderRefFormatException, self).__init__(
            customer_order_ref
        )
        self.customer_order_ref = customer_order_ref

    def __str__(self) -> str:
        return f"The supplied customer_order_ref doesnt meet the required length of {config.MIN_CUSTOMER_ORDER_REF_LENGTH}->{config.MAX_CUSTOMER_ORDER_REF_LENGTH}"


class BetRequestInvalidCustomerStrategyRefFormatException(BetConnectException):
    """
    raised when customer order ref is not between 1-15 character in length
    """

    def __init__(self, customer_strategy_ref: str):
        super(BetRequestInvalidCustomerStrategyRefFormatException, self).__init__(
            customer_strategy_ref
        )
        self.customer_strategy_ref = customer_strategy_ref

    def __str__(self) -> str:
        return f"The supplied customer_strategy_ref does not match the required length of {config.MIN_CUSTOMER_STRATEGY_REF_LENGTH}->{config.MAX_CUSTOMER_STRATEGY_REF_LENGTH}"


class APIError(BetConnectException):
    """
    Exception raised if error is found.
    """

    def __init__(
        self,
        response: Optional[dict],
        uri: str = None,
        params: dict = None,
        exception: Exception = None,
    ):
        if response:
            error_data = response.get("error")
            message = (
                "%s \nParams: %s \nException: %s \nError: %s \nFull Response: %s"
                % (uri, params, exception, error_data, response)
            )
        else:
            message = "%s \nParams: %s \nException: %s" % (uri, params, exception)
        super(APIError, self).__init__(message)



from betconnect.apiclient import APIClient
from decouple import config
from betconnect.enums import Environment
from betconnect import utils

import logging

logger = logging.getLogger(__name__)

"""
This file contains example of getting book percentage for all horse racing events
"""

# Create a trading client instance
client = APIClient(
    username=config("STAGING_BETCONNECT_USERNAME"),
    password=config("STAGING_BETCONNECT_PASSWORD"),
    api_key=config("STAGING_BETCONNECT_API_KEY"),
    environment=Environment.STAGING,
    personalised_production_url=config("PRODUCTION_URI"),
)

client.account.login()


# Active sports
active_sports = client.betting.active_sports()

try:
    horse_racing = [s for s in active_sports if s.display_name == "Horse Racing"][0]
except IndexError as e:
    raise Exception("Cannot find horse racing as an active sport!")

# Active bookmakers
active_bookmakers = client.betting.active_bookmakers()

# Active regions
active_regions = client.betting.active_regions(sport_id=horse_racing.sport_id)

try:
    active_region = [r for r in active_regions if r.name == "England"][0]
except IndexError:
    raise Exception("Cannot find the England region for horse racing!")

# Get active market type
market_types = client.betting.active_market_types(horse_racing.sport_id)

try:
    win_market_type = [t for t in market_types if t.name == "WIN"][0]
except IndexError:
    raise Exception("Cannot find the the market type WIN for the horse racing sport")


# Active fixtures
active_fixtures = client.betting.active_fixtures(
    sport_id=horse_racing.sport_id, region_id=active_region.region_id
)

book_percentages = []

for fixture in active_fixtures:
    selections_for_market = client.betting.selections_for_market(
        fixture_id=fixture.fixture_id, market_type_id=win_market_type.market_type_id
    )

    book_percentages.append(
        utils.calculate_book_percentage(market_selections=selections_for_market)
    )



from betconnect.apiclient import APIClient
from decouple import config
from betconnect.enums import Environment

import logging

logger = logging.getLogger(__name__)

"""
This file contains examples of account related requests
"""

# Create a trading client instance
client = APIClient(
    username=config("STAGING_BETCONNECT_USERNAME"),
    password=config("STAGING_BETCONNECT_PASSWORD"),
    api_key=config("STAGING_BETCONNECT_API_KEY"),
    environment=Environment.STAGING,
    personalised_production_url=config("PRODUCTION_URI"),
)

# login
login = client.account.login()

# get account balance
balance = client.account.get_balance()

# get account preferences
user_preferences = client.account.get_user_preferences()

# update client status (refresh token)
client.account.refresh_session_token()

# logout of client
client.account.logout()



from .__version__ import __title__, __version__, __author__
from .exceptions import (
    InValidPersonalisedProductionURL,
    UnknownBetConnectEnvironment,
    LoginMissingTokenInResponse,
    FailedLogin,
    MinOddException,
    UnexpectedResponseStatusCode,
    GamStopException,
    MissingUserPerferences,
    APIError,
    BetRequestIDFormatException,
    BetRequestIDStakeSizeException,
    BetRequestInvalidCustomerOrderRefFormatException,
    BetRequestInvalidCustomerStrategyRefFormatException,
)

from .enums import (
    Environment,
    BetStatus,
    BetSide,
    BetRequestStatus,
    MarketStatus,
    TradingStatus,
)

from .apiclient import APIClient



from betconnect.apiclient import APIClient
from betconnect import resources
from decouple import config
from betconnect.enums import Environment
import logging
from betconnect import enums
import uuid

logger = logging.getLogger(__name__)

"""
This script is an example workflow for betting on a horse racing market.
PLEASE MAKE SURE YOU ARE ONLY PRACTICING ORDERS AGAINST STAGING ENVIRONMENTS IF YOU HAVE NOT FULLY TESTED YOUR APP!
"""

# Strategy name max length = 15, no spaces. If you require longer use hashing function in utils.
STRATEGY_NAME = "horse_racing"

# Create a trading client instance
client = APIClient(
    username=config("STAGING_BETCONNECT_USERNAME"),
    password=config("STAGING_BETCONNECT_PASSWORD"),
    api_key=config("STAGING_BETCONNECT_API_KEY"),
    environment=Environment.STAGING,
    personalised_production_url=config("PRODUCTION_URI"),
)

assert client.environment == Environment.STAGING


# Login
login = client.account.login()

# Get account balance
balance = client.account.get_balance()

# Check there is available account balance
assert balance.balance > 0

# Get settled bets
my_back_bets = client.betting.my_bets(
    side=enums.BetSide.BACK, status=enums.BetRequestStatus.SETTLED
)

# Get active bets
my_active_back_bets = client.betting.my_bets(
    side=enums.BetSide.BACK, status=enums.BetRequestStatus.ACTIVE
)


# Active sports
active_sports = client.betting.active_sports()

try:
    horse_racing = [s for s in active_sports if s.display_name == "Horse Racing"][0]
except IndexError as e:
    raise Exception("Cannot find horse racing as an active sport!")

# Active bookmakers
active_bookmakers = client.betting.active_bookmakers()

# Active regions
active_regions = client.betting.active_regions(sport_id=horse_racing.sport_id)

try:
    active_region = [r for r in active_regions if r.name == "England"][0]
except IndexError:
    raise Exception("Cannot find the England region for horse racing!")

# Active fixtures
active_fixtures = client.betting.active_fixtures(
    sport_id=horse_racing.sport_id, region_id=active_region.region_id
)

# Get active market type
market_types = client.betting.active_market_types(horse_racing.sport_id)

try:
    win_market_type = [t for t in market_types if t.name == "WIN"][0]
except IndexError:
    raise Exception("Cannot find the the market type WIN for the horse racing sport")


# Get active selections and prices for active fixture
for fixture in active_fixtures[-1:]:

    # get active markets
    active_markets = client.betting.active_markets(fixture_id=fixture.fixture_id)

    if active_markets:
        try:
            win_bet_type = active_markets[0].bet_types[0]
        except IndexError:
            raise Exception(f"Cannot find any win types for the active market")

        # get the selections and prices for a given market
        fixture_selection_prices = client.betting.selections_for_market(
            fixture_id=fixture.fixture_id,
            market_type_id=win_market_type.market_type_id,
            top_price_only=False,
        )

        # get the top priced selections and prices for a given market
        top_price_fixture_selection_prices = client.betting.selections_for_market(
            fixture_id=fixture.fixture_id,
            market_type_id=win_market_type.market_type_id,
            top_price_only=True,
        )
        if fixture_selection_prices:
            # select the first selection/competitor with a price and that is active
            try:
                selection = [
                    c
                    for c in fixture_selection_prices
                    if (c.trading_status == "Trading") and c.prices
                ][0]
            except IndexError as e:
                raise Exception(
                    "No competitors in the fixture are active and have prices available"
                )

            try:
                best_price = [p for p in selection.prices][-1]
            except IndexError as e:
                raise Exception(
                    f"No price was available for selection {selection.name}"
                )

            # Create a BACK order against the first runner in the market
            bet_create_response = client.betting.bet_request_create(
                request_filter=resources.filters.CreateBetRequestFilter(
                    fixture_id=fixture.fixture_id,
                    market_type_id=win_market_type.market_type_id,
                    competitor=selection.competitor_id,
                    price=best_price.price,
                    stake=500,
                    bet_type=win_bet_type,
                    customer_strategy_ref=resources.CustomerStrategyRef.create_customer_strategy_ref(
                        STRATEGY_NAME
                    ),
                    customer_order_ref=resources.CustomerOrderRef.create_customer_order_ref(
                        str(uuid.uuid4())
                    ),
                )
            )
            if isinstance(bet_create_response, resources.BetRequestCreate):
                logger.info(f"Bet request was created!")

                # Retrieve my bets
                active_bet_request = client.betting.get_active_bet_requests()

                # Get active bets
                my_active_back_bets = client.betting.my_bets(
                    side=enums.BetSide.BACK, status=enums.BetRequestStatus.ACTIVE
                )

            else:
                logger.exception(f"Issue creating bet request with the supplied params")

        else:
            raise Exception(f"Could not find prices for fixture {fixture.fixture_id}")
    else:
        logger.info("No active markets could be found!")



from betconnect.apiclient import APIClient
from decouple import config
from betconnect.enums import Environment
from betconnect import resources
from betconnect import enums
from betconnect import utils

import logging

logger = logging.getLogger(__name__)

"""
This file contains examples of simple requests to BetConnect retrieve account data and betting history
"""

# Create a trading client instance
client = APIClient(
    username=config("STAGING_BETCONNECT_USERNAME"),
    password=config("STAGING_BETCONNECT_PASSWORD"),
    api_key=config("STAGING_BETCONNECT_API_KEY"),
    environment=Environment.STAGING,
    personalised_production_url=config("PRODUCTION_URI"),
)

# login
client.account.login()


# Get active bet requests
active_bet_requests = client.betting.get_active_bet_requests()

# Get my bets
my_bets = client.betting.my_bets(
    side=enums.BetSide.BACK,
    user_id=client.user_id,
    status=enums.BetRequestStatus.SETTLED,
    limit=100,
    page=0,
    get_all="get_all",
)


# Retrieve my bets
active_bet_request = client.betting.get_active_bet_requests()

# retrieves strategy based bets
for bets in active_bet_requests.bets:
    # Get my bets
    strategy_bet_requests = client.betting.my_bets(
        side=enums.BetSide.BACK,
        status=enums.BetRequestStatus.SETTLED,
        limit=100,
        page=0,
        get_all="get_all",
        customer_strategy_ref=bets.customer_strategy_ref,
    )

# Get historical bets
historical_bets = client.betting.bet_history(
    side=enums.BetSide.BACK, status=enums.BetRequestStatus.SETTLED
)



from betconnect.apiclient import APIClient
from betconnect import endpoints


class TestAPIClient:
    def test___init__(self):
        client = APIClient(
            username="test",
            password="123",
            api_key="456",
            personalised_production_url="https://jimbob.betconnect.com/",
        )
        assert client._username == "test"
        assert client._password == "123"
        assert client._api_key == "456"
        assert client._personalised_production_url == "https://jimbob.betconnect.com/"
        assert isinstance(client.account, endpoints.Account)
        assert isinstance(client.betting, endpoints.Betting)



import logging
import time
from typing import Union, Tuple, Optional
import requests
from betconnect import exceptions
from betconnect import resources
from .baseendpoint import BaseEndpoint

logger = logging.getLogger(__name__)


class Account(BaseEndpoint):
    def get_user_preferences(
        self,
    ) -> Union[resources.AccountPreferences, resources.BaseRequestException]:
        """
        Retrieves user preferences for the current account
        :return: A AccountPreferences resource or an BaseRequestException when BetConnect detects an issue.
        """

        (response, response_json, elapsed_time) = self._request(
            method_uri=f"{self.api_version}/get_user_preferences", authenticated=True
        )

        account_preferences = self.process_response(
            response=response,
            response_json=response_json,
            resource=resources.AccountPreferences,
            elapsed_time=elapsed_time,
        )
        if isinstance(account_preferences, resources.AccountPreferences):
            self.client.set_account_preferences(account_preferences=account_preferences)
        elif isinstance(account_preferences, resources.BaseRequestException):
            logger.warning(
                f"Issue trying to update the account balance. Request message: {account_preferences.message}"
            )
        return account_preferences

    def get_balance(self) -> Union[resources.Balance, resources.BaseRequestException]:
        """
        Gets the account balance
        :return: A Balance resource or an BaseRequestException when BetConnect detects an issue.
        """

        (response, response_json, elapsed_time) = self._request(
            method_uri=f"{self.api_version}/get_balance", authenticated=True
        )

        balance = self.process_response(
            response=response,
            response_json=response_json,
            resource=resources.Balance,
            elapsed_time=elapsed_time,
        )
        if isinstance(balance, resources.Balance):
            self.client.set_account_balance(account_balance=balance)
        elif isinstance(balance, resources.BaseRequestException):
            logger.warning(
                f"Issue trying to update the account balance. Request message: {balance.message}"
            )
        return balance

    def login(self) -> resources.Login:
        """
        Logs the user in with the username and password supplied to the client
        :return: Login resource
        """

        (response, response_json, elapsed_time) = self._post(
            method_uri=f"{self.api_version}/login"
        )

        data = response_json.get("data")

        if data:
            if "token" in data:
                self.client.process_login(data["token"])
                login_resource = resources.Login(**response_json)

                if self.client.account_preferences is None:
                    # try and get account preferences as well
                    account_preferences: Union[
                        resources.AccountPreferences, resources.BaseRequestException
                    ] = self.get_user_preferences()

                    if isinstance(account_preferences, resources.AccountPreferences):
                        self.client.set_account_preferences(
                            account_preferences=account_preferences
                        )
                    elif isinstance(
                        account_preferences, resources.BaseRequestException
                    ):
                        logger.warning(
                            f"Issue trying to update the account balance. Request message: {account_preferences.message}"
                        )

                if self.client.account_balance is None:
                    balance: Union[
                        resources.Balance, resources.BaseRequestException
                    ] = self.get_balance()

                    if isinstance(balance, resources.Balance):
                        self.client.set_account_balance(account_balance=balance)
                    elif isinstance(balance, resources.BaseRequestException):
                        logger.warning(
                            f"Issue trying to update the account balance. Request message: {balance.message}"
                        )

                return login_resource

            else:
                raise exceptions.LoginMissingTokenInResponse()
        else:
            raise exceptions.FailedLogin(username=self.client.username)

    def logout(self) -> None:
        """
        Logs the client out, removes all tokens.
        :return:
        """

        (response, response_json, elapsed_time) = self._post(
            method_uri=f"{self.api_version}/logout"
        )

        if response.status_code == 200:
            self.client.process_logout()
        else:
            logger.warning(
                f"There may have been an issue logging out. Expected status code 200, got status code {response.status_code}"
            )

    def refresh_session_token(self) -> Optional[resources.Login]:
        """
        Provides a refresh token for authentication
        :return: returns Login resource
        """

        (response, response_json, elapsed_time) = self._post(
            method_uri=f"{self.api_version}/status"
        )

        data = response_json.get("data")

        if data:
            if "refresh_token" in data:
                self.client.process_login(data["refresh_token"])
                return resources.Login(
                    **{
                        "data": {"token": data["refresh_token"]},
                        "message": response_json.get("message"),
                    }
                )
            else:
                logger.debug("No refresh token supplied")

    # noinspection PyProtectedMember
    def _post(
        self, method_uri: str, params: dict = None, authenticated: bool = True
    ) -> Tuple[requests.Response, dict, float]:
        """
        Post method
        :param method_uri: uri to be used, defined by each function.
        :param params: Query Params to be used in request
        :return: tuple of the Response, dict (json_dict), float (elapsed time)
        """
        params = params if params else {}

        uri = self.client.uri + method_uri
        time_sent = time.time()

        try:
            logger.debug(f"requesting the data for {uri}")

            response = self.session.post(uri, timeout=self._read_timeout)

        except requests.ConnectionError as e:
            raise exceptions.APIError(None, uri, params, e)
        except Exception as e:
            raise exceptions.APIError(None, uri, params, e)

        elapsed_time = time.time() - time_sent

        response_json = self.load_json_content(response)

        if self.check_status_code(response) is False:
            raise exceptions.UnexpectedResponseStatusCode(
                status_code=response.status_code, url=response.url
            )

        return response, response_json, elapsed_time

    def __call__(self, *args, **kwargs):
        return self.login()



import logging
from decouple import config
from betconnect import resources
from betconnect.resources import filters
from betconnect.apiclient import APIClient
from betconnect.enums import Environment

logger = logging.getLogger(__name__)

"""
This script is an example workflow for viewing active betrequest request in the market.
PLEASE MAKE SURE YOU ARE ONLY PRACTICING ORDERS AGAINST STAGING ENVIRONMENTS IF YOU HAVE NOT FULLY TESTED YOUR APP!
"""

# Create a trading client instance
client = APIClient(
    username=config("STAGING_BETCONNECT_USERNAME"),
    password=config("STAGING_BETCONNECT_PASSWORD"),
    api_key=config("STAGING_BETCONNECT_API_KEY"),
    environment=Environment.STAGING,
    personalised_production_url=config("PRODUCTION_URI"),
)

# login
client.account.login()

# Get active sports with bets
active_sports = client.betting.active_sports(with_bets=True)

market_bet_request = []

for active_sport in active_sports:
    if active_sport.bets_available > 0:

        bet_request_get = client.betting.bet_request_get(
            request_filter=filters.GetBetRequestFilter(sport_id=active_sport.sport_id)
        )
        if isinstance(bet_request_get, resources.BetRequest):
            market_bet_request.append(bet_request_get)



from betconnect.apiclient import APIClient
from decouple import config
from betconnect.enums import Environment

import logging

logger = logging.getLogger(__name__)

"""
This file contains examples of simple data requests to BetConnect
"""

# Create a trading client instance
client = APIClient(
    username=config("STAGING_BETCONNECT_USERNAME"),
    password=config("STAGING_BETCONNECT_PASSWORD"),
    api_key=config("STAGING_BETCONNECT_API_KEY"),
    environment=Environment.STAGING,
    personalised_production_url=config("PRODUCTION_URI"),
)

# login
client.account.login()


# get active book makers
active_bookmakers = client.betting.active_bookmakers()

# get active sports
active_sports = client.betting.active_sports()

# get the horse racing sport
try:
    horse_racing = [s for s in active_sports if s.name == "Horse Racing"][0]
except IndexError as e:
    raise Exception(f"No active sports could be found!")

# get active regions
active_regions = client.betting.active_regions(sport_id=horse_racing.sport_id)

# get the England region
try:
    england_region = [r for r in active_regions if r.name == "England"][0]
except IndexError as e:
    raise Exception(f"No England could be found. Poor England!")

# Get active market types for horse racing
active_market_types = client.betting.active_market_types(sport_id=horse_racing.sport_id)

try:
    win_market_type = [mt for mt in active_market_types if mt.name == "WIN"][0]
except IndexError as e:
    raise Exception("Could not find the WIN market type for horse racing")

# get active competitions for horse racing
active_competitions = client.betting.active_competitions(
    sport_id=horse_racing.sport_id, region_id=england_region.region_id
)

# get active horse racing fixtures
try:
    active_fixtures = client.betting.active_fixtures(
        sport_id=horse_racing.sport_id,
        region_id=england_region.region_id,
        competition_id=active_competitions[0].competition_id,
    )
except IndexError as e:
    raise Exception(
        f"Could not find any currently active horse racing competition. Check on the site they are currently available during the current hours."
    )


# get active markets for an active horse racing fixture
try:
    active_markets = client.betting.active_markets(
        fixture_id=active_fixtures[0].fixture_id
    )
    grouped_active_markets = client.betting.active_markets(
        fixture_id=active_fixtures[0].fixture_id, grouped=True
    )
except IndexError as e:
    raise Exception(f"Could not find any active fixtures for horse racing")

# get active selections for an active market
active_selections = client.betting.active_selections(
    fixture_id=active_fixtures[-1].fixture_id,
    market_type_id=win_market_type.market_type_id,
)

# get selections and prices for an active market
selections_for_market = client.betting.selections_for_market(
    fixture_id=active_fixtures[-1].fixture_id,
    market_type_id=win_market_type.market_type_id,
)

# Get prices
prices = client.betting.prices(
    fixture_id=active_fixtures[-1].fixture_id,
    market_type_id=win_market_type.market_type_id,
    competitor=selections_for_market[-1].competitor_id,
)



from datetime import datetime
from requests import Session
from betconnect.baseclient import BaseClient
from betconnect.enums import Environment
from betconnect import config


class TestBaseClient:
    def test___init__(self):
        client = BaseClient(
            username="test",
            password="123",
            api_key="456",
            environment=Environment.STAGING,
            personalised_production_url="https://jimbob.betconnect.com/",
        )
        assert isinstance(client.session, Session)
        assert client._username == "test"
        assert client._password == "123"
        assert client._api_key == "456"
        assert client._environment == Environment.STAGING
        assert client.session.headers["X-API-KEY"] == "456"
        assert isinstance(client.uri, str)
        assert client.logged_in is False
        assert client.login_date is None
        assert client.login_expiry_check is None
        assert client._token is None
        assert client.session_timeout == 28800

    def test__update_client_session(self, mock_base_client):
        mock_base_client._update_client_session()
        assert isinstance(mock_base_client.session, Session)
        assert mock_base_client.session.headers["X-API-KEY"] == "456"
        assert isinstance(mock_base_client.login_expiry_check, datetime)
        assert mock_base_client.session.auth == (
            mock_base_client._username,
            mock_base_client._password,
        )

        new_session = Session()
        new_session.auth = ("new_user", "new_password")
        mock_base_client._update_client_session(new_session)
        assert mock_base_client.session.auth == ("new_user", "new_password")

    def test__set_endpoint_uris(self, mock_base_client):
        mock_base_client._set_endpoint_uris(environment=Environment.STAGING)
        assert mock_base_client.uri == config.DEFAULT_STAGING_URL
        mock_base_client._set_endpoint_uris(environment=Environment.PRODUCTION)
        assert mock_base_client.uri == mock_base_client._personalised_production_url

    def test_process_login(self, mock_base_client):
        mock_base_client.process_login(token="test_token")
        assert isinstance(mock_base_client.login_date, datetime)
        assert isinstance(mock_base_client.login_expiry_check, datetime)
        assert mock_base_client._token == "test_token"
        assert mock_base_client.session.headers["X-AUTH-TOKEN"] == "test_token"



DEFAULT_STAGING_URL = "https://stgapi.betconnect.com/"
API_VERSION = "api/v2"
DEVELOPER_DOCS = "https://developer.betconnect.com/"
PAGE_START_VALUE = 1
MINIMUM_LIMIT_VALUE = 10

# TODO move these
MIN_CUSTOMER_STRATEGY_REF_LENGTH = 1
MAX_CUSTOMER_STRATEGY_REF_LENGTH = 15
MIN_CUSTOMER_ORDER_REF_LENGTH = 1
MAX_CUSTOMER_ORDER_REF_LENGTH = 36
CLIENT_READOUT_TIME_SECS = 120
LOGIN_FREQUENCY_CHECK_SECS = 60 * 120
CLIENT_TOKEN_REFRESH_FREQUENCY = 60 * 15
BET_REQUEST_MIN_ODDS = 1.01
SITE_MINIMUM_STAKE_SIZE = 1



from typing import Union, List
import requests
from betconnect import enums
from betconnect import resources
from .baseendpoint import BaseEndpoint
import logging
from betconnect import exceptions
from uuid import UUID

logger = logging.getLogger(__name__)


class Betting(BaseEndpoint):
    def active_bookmakers(self) -> List[resources.ActiveBookmaker]:
        """
        Returns a list of active sports
        :return: List of ActiveBookmaker
        """

        (response, response_json, elapsed_time) = self._request(
            method_uri=f"{self.api_version}/active_bookmakers", authenticated=True
        )

        return self.process_response(
            response=response,
            response_json=response_json,
            resource=resources.ActiveBookmaker,
            elapsed_time=elapsed_time,
        )

    def active_sports(self, with_bets: bool = False) -> List[resources.ActiveSport]:
        """
        Gets a list of active sports
        :param with_bets: boolean value (True, False). Returns the number of active bet_request available for each sport
        :return: List of ActiveSport
        """

        (response, response_json, elapsed_time) = self._request(
            method_uri=f"{self.api_version}/active_sports{'/True' if with_bets else ''}"
        )

        return self.process_response(
            response=response,
            response_json=response_json,
            resource=resources.ActiveSport,
            elapsed_time=elapsed_time,
        )

    def active_regions(self, sport_id: int) -> List[resources.ActiveRegion]:
        """
        Gets the active regions for a sport
        :param sport_id: The sport ID
        :return: List of ActiveRegion
        """

        (response, response_json, elapsed_time) = self._request(
            method_uri=f"{self.api_version}/active_regions/{sport_id}"
        )

        return self.process_response(
            response=response,
            response_json=response_json,
            resource=resources.ActiveRegion,
            elapsed_time=elapsed_time,
        )

    def active_competitions(
        self, sport_id: int, region_id: int
    ) -> List[resources.ActiveCompetition]:
        """
        Returns a list of active competitions for a given sport id and region id
        :param int sport_id: The sport ID
        :param int region_id: The region ID
        :return: List of ActiveCompetition
        """

        (response, response_json, elapsed_time) = self._request(
            method_uri=f"{self.api_version}/active_competitions/{sport_id}/{region_id}"
        )

        return self.process_response(
            response=response,
            response_json=response_json,
            resource=resources.ActiveCompetition,
            elapsed_time=elapsed_time,
        )

    def active_fixtures(
        self, sport_id: int, region_id: int = None, competition_id: int = None
    ) -> List[resources.ActiveFixture]:
        """
        Gets active fixtures for a sport, region and competition
        :param sport_id: The sport ID
        :param region_id: The region ID
        :param competition_id: The competition ID
        :return: List of ActiveFixture
        """
        if competition_id:
            assert region_id is not None

        (response, response_json, elapsed_time) = self._request(
            method_uri=f"{self.api_version}/active_fixtures/{sport_id}/{region_id if region_id else ''}{f'/{competition_id}' if competition_id is not None else ''}"
        )

        return self.process_response(
            response=response,
            response_json=response_json,
            resource=resources.ActiveFixture,
            elapsed_time=elapsed_time,
        )

    def active_market_types(self, sport_id: int) -> List[resources.ActiveMarketType]:
        """
        Gets the active market type for a sport
        :param sport_id: The sport ID
        :return: List of ActiveMarketType
        """

        (response, response_json, elapsed_time) = self._request(
            method_uri=f"{self.api_version}/active_market_types/{sport_id}"
        )

        return self.process_response(
            response=response,
            response_json=response_json,
            resource=resources.ActiveMarketType,
            elapsed_time=elapsed_time,
        )

    def active_markets(
        self, fixture_id: int, grouped: bool = False
    ) -> List[resources.ActiveMarket]:
        """
        Gets a list of active markets for the fixture
        :param fixture_id: The fixture ID
        :param grouped: boolean value to determine whether the response should be grouped
        :return: List of ActiveMarket
        """

        (response, response_json, elapsed_time) = self._request(
            method_uri=f"{self.api_version}/active_markets/{fixture_id}{'/grouped' if grouped else ''}"
        )

        return self.process_response(
            response=response,
            response_json=response_json,
            resource=resources.ActiveMarket,
            elapsed_time=elapsed_time,
        )

    def active_selections(
        self, fixture_id: int, market_type_id: int, handicap: str = None
    ) -> List[resources.ActiveSelection]:
        """
        Gets active selections for a given fixture and market type
        :param fixture_id: The fixture ID
        :param market_type_id: The market type ID
        :param handicap: string handicap value. example '4.5'
        :return: List of ActiveSelection
        """
        (response, response_json, elapsed_time) = self._request(
            method_uri=f"{self.api_version}/active_selections/{fixture_id}/{market_type_id}{f'/{handicap}' if handicap is not None else ''}"
        )

        return self.process_response(
            response=response,
            response_json=response_json,
            resource=resources.ActiveSelection,
            elapsed_time=elapsed_time,
        )

    def bet_request_create(
        self, request_filter: resources.CreateBetRequestFilter
    ) -> Union[resources.BaseRequestException, resources.BetRequestCreate]:
        """
        Creates a bet (Authenticated)
        :param request_filter: The bet create filter
        :return: a successful BetRequestCreate resource or an BaseRequestException when BetConnect detects an issue.
        """

        (response, response_json, elapsed_time) = self._post(
            method_uri=f"{self.api_version}/bet_request_create",
            data=request_filter.generate_request_data(exclude_none=True),
        )

        return self.process_response(
            response=response,
            response_json=response_json,
            resource=resources.BetRequestCreate,
            elapsed_time=elapsed_time,
        )

    def bet_request_get(
        self, request_filter: resources.GetBetRequestFilter
    ) -> Union[resources.BaseRequestException, resources.BetRequest]:
        """
        Gets a bet request for the given filter.

        This is other users bet requests (please use get_active_bet_requests / my_bets ) to get your own.

        :param request_filter: A bet request filter, either enter a bet request id or other filter values
        :return: A bet BetRequest resource or an BaseRequestException when BetConnect detects an issue.
        """

        if request_filter.bet_request_id:
            (response, response_json, elapsed_time) = self._post(
                method_uri=f"{self.api_version}/bet_request_get",
                data=request_filter.generate_request_data(
                    exclude={
                        "sport_id",
                        "bookmakers",
                        "min_odds",
                        "max_odds",
                        "accept_each_way",
                    }
                ),
            )
        else:
            if request_filter.accept_each_way:
                (response, response_json, elapsed_time) = self._post(
                    method_uri=f"{self.api_version}/bet_request_get",
                    data=request_filter.generate_request_data(
                        exclude={"bet_request_id", "accept_each_way"}
                    ),
                )
            else:
                (response, response_json, elapsed_time) = self._post(
                    method_uri=f"{self.api_version}/bet_request_get",
                    data=request_filter.generate_request_data(
                        exclude={"bet_request_id"}
                    ),
                )

        return self.process_response(
            response=response,
            response_json=response_json,
            resource=resources.BetRequest,
            elapsed_time=elapsed_time,
        )

    def _is_line_market(self, response: requests.Response, response_json: dict) -> bool:
        """
        Function to check whether the market is a line market returned from BetConnect
        :param response: A response resource return from BetConnect
        :param response_json: The response json containing thee data
        :return: A boolean {True, False} as whether it is a Line market
        """
        if self.check_status_code(response):
            if "data" in response_json:
                data = response_json["data"]

                if isinstance(data, list):
                    if data:
                        return True if "line" in data[0] else False

        return False

    def selections_for_market(
        self, fixture_id: int, market_type_id: int, top_price_only: bool = False
    ) -> Union[resources.BaseRequestException, List[resources.SelectionsForMarket]]:
        """
        Returns the selection for a given market with their prices. This method is the fastest way to get selections
        and prices in few requests.
        :param fixture_id: The fixture ID
        :param market_type_id: the market type ID
        :param top_price_only: a bool value that represents whether to only return the top value
        :return: A List of SelectionsForMarket resources or an BaseRequestException when BetConnect detects an issue.
        """
        (response, response_json, elapsed_time) = self._request(
            method_uri=f"{self.api_version}/selections_for_market/{fixture_id}/{market_type_id}/{f'True'if top_price_only else 'False'}"
        )
        # Check is response a line market
        if self._is_line_market(response=response, response_json=response_json):
            return self.process_response(
                response=response,
                response_json=response_json,
                resource=resources.LineMarketsSelectionsForMarket,
                elapsed_time=elapsed_time,
            )
        else:
            return self.process_response(
                response=response,
                response_json=response_json,
                resource=resources.SelectionsForMarket,
                elapsed_time=elapsed_time,
            )

    def bet_request_match(
        self, bet_request_id: UUID, accepted_stake: int
    ) -> Union[resources.BaseRequestException, resources.BetRequestMatch]:
        """
        A request to match for part or all of an active bet request.
        :param bet_request_id: The bet request ID
        :param accepted_stake: the target stake you would like to match
        :return: A BetRequestMatch resource or an BaseRequestException when BetConnect detects an issue.
        """
        self.check_bet_request_id(bet_request_id=bet_request_id)

        (response, response_json, elapsed_time) = self._patch(
            method_uri=f"{self.api_version}/bet_request_match",
            data={
                "bet_request_id": str(bet_request_id),
                "accepted_stake": accepted_stake,
            },
        )

        return self.process_response(
            response=response,
            response_json=response_json,
            resource=resources.BetRequestMatch,
            elapsed_time=elapsed_time,
        )

    def bet_request_match_more(
        self, bet_request_id: UUID, requested_stake: float
    ) -> Union[resources.BetRequestMatchMore, resources.BaseRequestException]:
        """
        Allows you to match a bet request fully (up to the maximum liability allowed for a single bet)
        and request more of that bet until no liability is left.
        :param bet_request_id: The bet request ID
        :param requested_stake: The requested stake
        :return: A BetRequestMatchMore resource or an BaseRequestException when BetConnect detects an issue.
        """
        self.check_bet_request_id(bet_request_id=bet_request_id)

        (response, response_json, elapsed_time) = self._patch(
            method_uri=f"{self.api_version}/bet_request_match_more",
            data={
                "bet_request_id": str(bet_request_id),
                "requested_stake": int(requested_stake),
            },
        )

        return self.process_response(
            response=response,
            response_json=response_json,
            resource=resources.BetRequestMatchMore,
            elapsed_time=elapsed_time,
        )

    def bet_request_stop(
        self, bet_request_id: UUID, stop_bet_reason: str = ""
    ) -> Union[resources.BetRequestStop, resources.BaseRequestException]:
        """
        Stops an active bet request
        :param bet_request_id: The bet request ID
        :param stop_bet_reason: optional reason (i.e. 'no longer value')
        :return: A BetRequestStop resource or an BaseRequestException when BetConnect detects an issue.
        """
        self.check_bet_request_id(bet_request_id=bet_request_id)

        (response, response_json, elapsed_time) = self._post(
            method_uri=f"{self.api_version}/bet_request_stop",
            data={
                "bet_request_id": str(bet_request_id),
                "stop_bet_reason": stop_bet_reason,
            },
            authenticated=True,
        )

        return self.process_response(
            response=response,
            response_json=response_json,
            resource=resources.BetRequestStop,
            elapsed_time=elapsed_time,
        )

    def get_active_bet_requests(
        self, limit: int = None, page: int = None
    ) -> Union[resources.ActiveBetRequests, resources.BaseRequestException]:
        """
        Gets active bet requests, taking into account pagination
        :param limit: Limit the number of active bets returned
        :param page: The page starting number
        :return: A ActiveBetRequests resource or an BaseRequestException when BetConnect detects an issue.
        """
        if (page is not None) or (limit is not None):
            limit = (
                max(limit, self.client.minimum_limit_value)
                if limit
                else self.client.minimum_limit_value
            )
            page = (
                max(page, self.client.page_start_value)
                if page
                else self.client.page_start_value
            )

        (response, response_json, elapsed_time) = self._request(
            method_uri=f"{self.api_version}/get_active_bet_requests{f'/{limit}/{page}' if page is not None else ''}",
            authenticated=True,
        )

        return self.process_response(
            response=response,
            response_json=response_json,
            resource=resources.ActiveBetRequests,
            elapsed_time=elapsed_time,
        )

    def prices(
        self,
        fixture_id: int,
        market_type_id: int,
        competitor: str,
        handicap: str = None,
    ) -> List[resources.Price]:
        """
        Returns a set of prices for a given fixture, market type and competition.
        :param fixture_id: The fixture ID
        :param market_type_id: the market type ID
        :param competitor: the compeitor string
        :param handicap: handicap string value (If the market type is a handicap market a handicap is required)
        :return: A List of Price resources
        """

        (response, response_json, elapsed_time) = self._request(
            method_uri=f"{self.api_version}/prices/{fixture_id}/{market_type_id}/{competitor.lower()}{f'/{handicap}' if handicap is not None else ''}"
        )

        return self.process_response(
            response=response,
            response_json=response_json,
            resource=resources.Price,
            elapsed_time=elapsed_time,
        )

    def bet_history(
        self,
        status: enums.BetStatus,
        side: enums.BetSide,
        limit: int = None,
        page: int = None,
    ) -> Union[resources.BaseRequestException, resources.BetHistoryRequest]:
        """
        Returns bet history with paging options
        :param status: The status of the bet
        :param side: BetSide enum - either back or lay
        :param limit: Limit the number of bets returned
        :param page: The page number to start from
        :return: A BetHistoryRequest resource or an BaseRequestException when BetConnect detects an issue.
        """
        if (page is not None) or (limit is not None):
            limit = (
                max(limit, self.client.minimum_limit_value)
                if limit
                else self.client.minimum_limit_value
            )
            page = (
                max(page, self.client.page_start_value)
                if page
                else self.client.page_start_value
            )

        (response, response_json, elapsed_time) = self._request(
            method_uri=f"{self.api_version}/bet_history/{str(side.value)}/{self.client.username}/{str(status.value)}{f'/{limit}/{page}'if page is not None else ''}",
            authenticated=True,
        )

        return self.process_response(
            response=response,
            response_json=response_json,
            resource=resources.BetHistoryRequest,
            elapsed_time=elapsed_time,
        )

    def get_viewed_next_page(
        self, bet_request_id: UUID, sport_id: int = None
    ) -> Union[resources.Viewed, resources.BaseRequestException]:
        """
        Returns the next bet request available to be viewed. (mprefered approach is to use bet_request_get
        with a wide filter instead)
        :param bet_request_id: The bet request ID
        :param sport_id: the sport ID
        :return: A Viewed resource or an BaseRequestException when BetConnect detects an issue.
        """
        self.check_bet_request_id(bet_request_id=bet_request_id)

        (response, response_json, elapsed_time) = self._request(
            method_uri=f"{self.api_version}/get_viewed_next_prev/{str(bet_request_id)}{f'/{sport_id}'if sport_id is not None else ''}"
        )

        return self.process_response(
            response=response,
            response_json=response_json,
            resource=resources.Viewed,
            elapsed_time=elapsed_time,
        )

    def my_bets(
        self,
        side: enums.BetSide,
        status: enums.BetRequestStatus,
        user_id: str = None,
        limit: int = None,
        page: int = None,
        get_all: str = None,
        customer_strategy_ref: str = None,
    ) -> Union[
        resources.MyBetsBetRequests,
        resources.MyBetsBets,
        resources.BaseRequestException,
    ]:
        """
        Returns a list of MyBets allowing for per strategy recall of bets
        :param side: the side of the bet, enum value BACK and LAY
        :param status: The bet request status, enum values ACTIVE and SETTLED
        :param user_id: The user_id, optional will try to get this from the user preference which is called on login.
        :param limit: Limit the number of bets coming back.
        :param page: T page number for (pagination)
        :param get_all: string value to return all values with the above limit and page
        :param customer_strategy_ref: The customer strategy ref which has been attached to any bet requests
        :return: A MyBetsBetRequests resource or an BaseRequestException when BetConnect detects an issue.
        """

        if user_id is None:
            if self.client.user_id:
                user_id = self.client.user_id
            else:
                logger.debug(
                    f"Trying to get your user preferences as no user_id was supplied!"
                )
                self.client.account.get_user_preferences()
                if self.client.user_id:
                    user_id = self.client.user_id
                else:
                    raise exceptions.MissingUserPerferences()

        uri_extension = ""

        if (
            (page is not None)
            or (limit is not None)
            or (get_all is not None)
            or (customer_strategy_ref is not None)
        ):

            limit = (
                max(limit, self.client.minimum_limit_value)
                if limit is not None
                else self.client.minimum_limit_value
            )
            page = (
                max(page, self.client.page_start_value)
                if page is not None
                else self.client.page_start_value
            )
            get_all = "get_all"

            uri_extension += f"/{limit}/{page}/{get_all}"

            uri_extension += (
                f"/{customer_strategy_ref}" if customer_strategy_ref else ""
            )

        (response, response_json, elapsed_time) = self._request(
            method_uri=f"{self.api_version}/my_bets/{side.value}/{user_id}/{str(status.value)}{uri_extension}",
            authenticated=True,
        )

        if side.value == "back":
            return self.process_response(
                response=response,
                response_json=response_json,
                resource=resources.MyBetsBetRequests,
                elapsed_time=elapsed_time,
            )
        else:
            return self.process_response(
                response=response,
                response_json=response_json,
                resource=resources.MyBetsBets,
                elapsed_time=elapsed_time,
            )

    def lock_bet(
        self, bet_request_id: UUID, bet_status_id: int, allocated_stake: int
    ) -> Union[resources.LockBet, resources.BaseRequestException]:
        """
        Matched betting premium product only. Premuim subscription required.
        client.account_preferences.is_premium_subscriber ==1.
        :param bet_request_id: The bet request ID
        :param bet_status_id: The status ID of the bet
        :param allocated_stake:
        :return: LockBet
        """
        raise NotImplementedError

        (response, response_json, elapsed_time) = self._put(
            method_uri=f"{self.api_version}/lock_bet",
            data={
                "bet_request_id": str(bet_request_id),
                "bet_status_id": bet_status_id,
                "allocated_stake": allocated_stake,
            },
        )
        return self.process_response(
            response=response,
            response_json=response_json,
            resource=resources.LockBet,
            elapsed_time=elapsed_time,
        )
